(*
Copyright 2011 Christian Sternagel, René Thiemann

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory Matchbounds_Impl 
imports
  Matchbounds
  Tree_Automata_Impl
  Trs_Impl 
  Signature_Extension
  QDP_Framework_Impl
begin

lemma [code]:
  "Matchbounds.roof (l, r) = (
    let xs = vars_term_list r in (\<lambda>t.
      let xt = vars_term_list t in
      (\<forall>x\<in>set xs. x \<in> set xt)))"
  unfolding Let_def roof.simps by (rule ext, simp) blast

lemma e_bounds:
  fixes R :: "('f, 'q) trs"
    and TA :: "('q, 'f \<times> nat) ta"
    and L :: "('f, 'q) term set"
  assumes 
      ll: "left_linear_trs R"
    and SN: "locally_terminating (cover e R)"
    and wf_trs: "wf_trs R"
    and fin: "finite (ta_rules TA)"
    and compat: "ta_trs_compatible TA (cover e R)"
    and incl: "(lift_term 0) ` L \<subseteq> ta_lang TA"
    and F: "finite F"
    and R: "finite R"
    and L: "\<Union> funas_term ` L \<subseteq> F"
    and t: "t \<in> L"
  shows "SN_on (rstep R) {t}"
proof -
  have compact: "compact (lift_term 0 ` L) (cover e R)"
    by (rule ta_compact[OF fin compat cover_left_linear[OF ll] cover_var_condition incl], insert wf_trs[unfolded wf_trs_def], auto)
  from this[unfolded compact_def] obtain G where fin: "finite G"
    and G: "\<Union> funas_term ` {u. \<exists> s t. u = t \<and> s \<in> lift_term 0 ` L \<and> (s,t) \<in> (rstep (cover e R))^*} \<subseteq> G" by auto
  from finite_list[OF fin] obtain Gl where fin: "G = set Gl" by auto
  let ?h = "(\<lambda>(f,n). height f)"
  obtain c where c: "c = max_list (map ?h Gl)" by auto
  have bounded: "e_bounded e c R L"
    unfolding e_bounded_def
  proof (intro allI impI)
    fix s t f n
    assume "s \<in> lift_term 0 ` L \<and> (s,t) \<in> (rstep (cover e R))^* \<and> (f,n) \<in> funas_term t"
    hence "(f,n) \<in> funas_term t" and "funas_term t \<subseteq> set Gl" using G unfolding fin by auto
    hence "(f,n) \<in> set Gl" by auto
    hence "height f \<in> set (map ?h Gl)" by auto
    from max_list[OF this]
    show "height f \<le> c" unfolding c .
  qed
  from e_bounded[OF wf_trs SN R F L ll bounded t]
  show "SN_on (rstep R) {t}" .
qed

definition cover_bound where 
  "cover_bound e c R \<equiv> {(l,r). (l,r) \<in> cover e R \<and> (\<forall> f n. (f,n) \<in> funas_term l \<longrightarrow> height f \<le> c)}"

lemma e_bounds_impl: fixes R :: "('f,'q)rules"
  and TA :: "('q,'f \<times> nat)ta"
  and qfin :: "'q set"
  and c :: nat
  assumes 
      ll: "left_linear_trs (set R)"
  and wf_trs: "wf_trs (set R)"
  and fin: "finite (ta_rules TA)"
  and SN: "locally_terminating (cover e (set R))"
  and compat: "ta_trs_compatible TA (cover_bound e c (set R))" (is "ta_trs_compatible _ ?R")
  and c: "\<And> f n. (f,n) \<in> ta_syms TA \<Longrightarrow> height f \<le> c"
  and contain: "ta_contains ((\<lambda>(f,n). (lift 0 f,n)) ` set (funas_trs_list R)) TA qfin" (is "ta_contains (?lz ` _) _ _")
  and qfin: "\<exists> q \<in> qfin. q \<in> ta_rqs TA"
  shows "SN (rstep (set R))"
proof -
  from qfin obtain q where q: "q \<in> qfin \<and> q \<in> ta_rqs TA" by auto
  let ?const = "SOME c. (c,0) \<in> funas_trs (set R)"
  obtain const where const_def: "const = ?const" by auto
  let ?cond = "(const,0) \<in> funas_trs (set R)"
  let ?TA = "\<lparr> ta_final = (ta_final TA), ta_rules = insert ((const,0) [] \<rightarrow> q) (ta_rules TA), ta_eps = ta_eps TA  \<rparr>"
  let ?TA' = "if ?cond then TA else ?TA"
  obtain TA' where TA': "TA' = ?TA'" by auto
  let ?F = "(const,0) # funas_trs_list R"
  obtain F where F: "F = ?F" by simp
  have const: "(const,0) \<in> set F" by (simp add: F)
  have add_constant: "ta_trs_compatible TA' (cover_bound e c (set R)) \<and> ta_contains (?lz ` set F) TA' qfin" (is "?comp \<and> ?cont")
  proof (cases ?cond)
    case True
    have "set (map ?lz F) = set (map ?lz (funas_trs_list R))" unfolding F 
      using True by auto
    hence "ta_contains (?lz ` set F) TA qfin" using contain unfolding ta_contains_def 
      by auto
    with compat True show ?thesis unfolding TA' by auto
  next
    case False
    hence TA': "TA' = ?TA" unfolding TA' by auto
    from contain have qfin:"qfin \<subseteq> ta_final ?TA" unfolding ta_contains_def by auto
    {
      fix f n 
      assume fn: "(f,n) \<in> ?lz ` set F"
      hence "\<forall> qs'. length qs' = n \<and> set qs' \<subseteq> qfin \<longrightarrow> (\<exists> q q'. (f qs' \<rightarrow> q) \<in> ta_rules ?TA \<and> q' \<in> qfin \<and> (q,q') \<in> (ta_eps ?TA)^*)"
      proof (cases "(f,n) = ((const,0),0)")
        case True
        thus ?thesis using q  by auto
      next
        case False
        with fn have fn: "(f,n) \<in> ?lz ` set (funas_trs_list R)" unfolding F by auto
        hence fn: "(f,n) \<in> ?lz ` (funas_trs (set R))" by auto
        then obtain g where f: "f = (g,0)" and g: "(g,n) \<in> funas_trs (set R)" by auto
        show ?thesis
        proof (intro allI impI)
          fix qs'
          assume len: "length qs' = n \<and> set qs' \<subseteq> qfin"
          from contain[unfolded ta_contains_def, THEN conjunct2, simplified,
            THEN spec[of _ g], THEN spec[of _ 0], THEN spec[of _ qs']] 
          fn[unfolded f] g len 
          show "\<exists> q q'. f qs' \<rightarrow> q \<in> ta_rules ?TA \<and> q' \<in> qfin \<and> (q,q') \<in> (ta_eps ?TA)^*"
            unfolding f by auto
        qed
      qed
    } note main = this
    from main qfin
    have contain: "ta_contains (?lz ` set F) ?TA qfin" unfolding ta_contains_def ta.simps
      by blast
    have compat: "ta_trs_compatible ?TA (cover_bound e c (set R))"
      unfolding ta_trs_compatible_def
    proof (clarify)
      fix l r
      assume lr: "(l,r) \<in> cover_bound e c (set R)"
        and lsyms: "funas_term l \<subseteq> ta_syms ?TA"
      from lsyms have lsyms: "funas_term l \<subseteq> insert ((const,0),0) (ta_syms TA)" unfolding ta_syms_def by auto
      from lr obtain ll rr where llrr:"(ll,rr) \<in> set R" and base:"ll = base_term l"
        unfolding cover_bound_def cover_def by auto
      {
        assume "((const,0),0) \<in> funas_term l" 
        hence "base ((const,0),0) \<in> funas_term ll" unfolding base 
          unfolding funas_term_map_term by force
        hence "(const,0) \<in> funas_term ll" by auto
        with llrr False have False unfolding funas_trs_def funas_rule_def by auto
      } hence const_new: "((const,0),0) \<notin> funas_term l" ..
      with lsyms have lsyms: "funas_term l \<subseteq> ta_syms TA" by auto
      from compat[unfolded ta_trs_compatible_def] lr lsyms have 
        compat: "ta_rule_compatible TA (l,r)" by auto      
      {
        fix q'
        assume q': "(q,q') \<in> (ta_eps TA)^*"
        from q[unfolded ta_rqs_def] obtain q'' rule where
          q: "(q'',q) \<in> (ta_eps TA)^*" and rule: "rule \<in> ta_rules TA" and sym: "rq rule = q''" by auto
        from q' q have q': "(q'',q') \<in> (ta_eps TA)^*" by auto
        with rule sym 
        have "q' \<in> ta_rqs TA" unfolding ta_rqs_def by blast
      }        
      hence ta_rqs: "ta_rqs ?TA = ta_rqs TA" unfolding ta_rqs_def by auto
      show "ta_rule_compatible ?TA (l,r)" 
        unfolding ta_rule_compatible_def ta_rqs
      proof (rule, intro allI impI)
        fix \<tau>
        assume \<tau>: "\<tau> ` vars_term l \<subseteq> ta_rqs TA"
        obtain rr where rr: "map_vars \<tau> r = rr" by auto
        show "ta_res ?TA (map_vars \<tau> l) \<subseteq> ta_res ?TA (map_vars \<tau> r)"
        proof
          fix q'
          assume "q' \<in> ta_res ?TA (map_vars \<tau> l)"
          hence "q' \<in> ta_res TA (map_vars \<tau> l)" using const_new
          proof (induct l arbitrary: q')
            case (Var x) thus ?case by auto
          next
            case (Fun f ls)
            from Fun(3) have f: "(f,length ls) \<noteq> ((const,0),0)" by auto
            from Fun(2)[unfolded map_vars.simps ta_res.simps]
            obtain q2 qs where rule: "f qs \<rightarrow> q2 \<in> ta_rules ?TA" and len: "length qs = length (map (map_vars \<tau>) ls)"
              and ind: "\<forall>i < length ls. qs ! i \<in> map (ta_res ?TA) (map (map_vars \<tau>) ls) ! i" 
              and q2: "(q2,q') \<in> (ta_eps TA)^*" by auto
            from rule len f have rule: "f qs \<rightarrow> q2 \<in> ta_rules TA" by auto
            show ?case
            proof (unfold map_vars.simps ta_res.simps, rule, intro exI conjI, rule HOL.refl, rule rule, rule q2,
              rule len, intro allI impI)
              fix i
              assume i: "i < length (map (map_vars \<tau>) ls)"
              hence i': "i < length ls" by auto
              from i' have mem: "ls ! i \<in> set ls" by auto
              with Fun(3) have nmem: "((const,0),0) \<notin> funas_term (ls ! i)" by auto
              show "qs ! i \<in> map (ta_res TA) (map (map_vars \<tau>) ls) ! i"
                unfolding nth_map[OF i] nth_map[OF i']
                by (rule Fun(1)[OF mem _ nmem], insert ind i', auto)
            qed
          qed                        
          with compat[unfolded ta_rule_compatible_def] \<tau>
          have "q' \<in> ta_res TA (map_vars \<tau> r)" by auto
          thus "q' \<in> ta_res ?TA (map_vars \<tau> r)" unfolding rr
          proof (induct rr arbitrary: q')
            case (Var q) thus ?case by auto
          next
            case (Fun f ts)
            from Fun(2) obtain q2 qs where rule: "f qs \<rightarrow> q2 \<in> ta_rules ?TA"
              and len: "length qs = length ts"
              and rec: "\<And>i. i < length ts \<Longrightarrow> qs ! i \<in> ta_res TA (ts ! i)"
              and q: "(q2,q') \<in> (ta_eps TA)^*"
              by auto
            {
              fix i
              assume i: "i < length ts"
              hence "ts ! i \<in> set ts" by auto
              from Fun(1)[OF this rec[OF i]]
              have "qs ! i \<in> ta_res ?TA (ts ! i)" by auto
            }
            with rule len q show ?case by auto
          qed
        qed
      qed
    qed
    thus ?thesis using contain unfolding TA' by auto
  qed
  hence compat: ?comp and contain: ?cont by auto
  have subF: "funas_trs (set R) \<subseteq> set F" unfolding F funas_trs_list_sound by auto
  have finF: "finite (set F)" by auto
  have finR: "finite (set R)" by auto
  let ?wF = "{t . funas_term t \<subseteq> set F} :: ('f,'q)term set"
  let ?L = "{t. t \<in> ?wF \<and> ground t}"  
  have L: "?L \<subseteq> ?wF" by auto
  show ?thesis
  proof (rule SN_wf_ground[OF const subF])
    fix t :: "('f,'q)term"
    assume wf: "funas_term t \<subseteq> set F" and gr: "ground t"
    hence t: "t \<in> ?L" by auto
    show "SN_on (rstep (set R)) {t}"
    proof (rule e_bounds[OF ll SN wf_trs _ _ _ finF finR _ t])
      show "lift_term 0 ` {t \<in> ?wF. ground t} \<subseteq> ta_lang TA'" (is "?L \<subseteq> _")
      proof (rule subset_trans[OF _ ta_contains[OF contain[unfolded F[symmetric]]]])
        let ?F = "?lz ` set F"
        let ?G = "{t. ground t}"
        let ?wLF = "{t. funas_term t \<subseteq> ?F} :: ('f \<times> nat,'q)term set"
        let ?R = "?wLF \<inter> ?G"
        show "?L \<subseteq> ?R" 
        proof
          fix t
          assume "t \<in> ?L"
          then obtain s where wf: "s \<in> ?wF" and s: "ground s"
            and t: "t = lift_term 0 s" by auto
          show "t \<in> ?R" unfolding t using wf s
          proof (induct s)
            case (Fun f ss)
            {
              fix s
              assume s: "s \<in> set ss"
              from Fun(3) s have g: "ground s" by auto
              from Fun(2) s have wf: "s \<in> ?wF" by auto
              from Fun(1)[OF s wf g] have l: "lift_term 0 s \<in> ?R" .
              from Fun(1)[OF s wf g] have g: "ground (lift_term 0 s)" by blast
              note g l
            } note ind = this
            from ind(1) have g: "ground (lift_term 0 (Fun f ss))" by auto
            have l: "lift_term 0 (Fun f ss) \<in> ?wLF"
              using ind(2) Fun(2) by auto
            show ?case
              by (rule, rule l, rule, rule g)
          qed simp
        qed
      qed
    next
      {
        fix l r 
        assume rule: "(l,r) \<in> cover e (set R)" and syms: "funas_term l \<subseteq> ta_syms TA'"
        from c have c: "\<And> f n. (f,n) \<in> ta_syms TA' \<Longrightarrow> height f \<le> c" unfolding TA' 
          by (cases ?cond, auto simp: ta_syms_def, blast)
        have "(l,r) \<in> ?R" using syms rule c unfolding cover_bound_def by auto 
        with compat[unfolded ta_trs_compatible_def] syms 
        have "ta_rule_compatible TA' (l,r)" by auto
      }        
      thus "ta_trs_compatible TA' (cover e (set R))"
        unfolding ta_trs_compatible_def by auto
    next
      show "finite (ta_rules TA')"
        unfolding TA' using fin by auto
    qed auto
  qed
qed


fun flatten_term_enum_filter :: "(('f,'v)term \<Rightarrow> bool) \<Rightarrow> ('f list,'v)term \<Rightarrow> ('f,'v)term list"
where "flatten_term_enum_filter f (Var x) = (let tx = Var x in if f tx then [tx] else [])"
   |  "flatten_term_enum_filter f (Fun fs ts) = (
          let lts = map (flatten_term_enum_filter f) ts 
            in (if list_ex (\<lambda> ts. ts = []) lts then [] else 
              (let ss = concat_lists lts
            in filter f (concat (map (\<lambda> f. map (Fun f) ss) fs)))))"

  
lemma flatten_term_enum_filter: 
  shows "set (flatten_term_enum_filter f t) = {u. instance_term u (map_term set t) \<and> (\<forall> v \<unlhd> u. f v)}"
proof (induct t)
  case (Var x)
  show ?case (is "_ = ?R")
  proof -
    {
      fix t 
      assume "t \<in> ?R"
      hence "t = Var x" by (cases t, auto)
    } 
    thus ?thesis using supteq_Var_id[of x] by (auto simp: Let_def)
  qed
next
  case (Fun fs ts)
  show ?case (is "?L = ?R")
  proof -
    {
      fix i
      assume "i < length ts"
      hence "ts ! i \<in> set ts" by auto
      note Fun[OF this]
    } note ind = this
    have idL: "?L = {Fun g ss | g ss. g \<in> set fs  \<and> length ss = length ts \<and> f (Fun g ss) \<and> (\<forall>i<length ts. ss ! i \<in> set (flatten_term_enum_filter f (ts ! i)))}" (is "_ = ?M1") by (simp add: Let_def list_ex_iff set_conv_nth[of ts], auto)
    let ?R1 = "{Fun g ss | g ss. g \<in> set fs \<and> length ss = length ts \<and> f (Fun g ss) \<and> (\<forall> i<length ss. instance_term (ss ! i) (map_term set (ts ! i)) \<and> (\<forall> u \<unlhd> (ss ! i). f u))}"
    {
      fix u
      have "(u \<in> ?R) = (u \<in> ?R1)" 
      proof (cases u)
        case (Fun g ss)
        show ?thesis unfolding Fun 
          by (auto simp: Fun_supteq set_conv_nth[of ss])
      qed auto        
    }
    hence idR: "?R = ?R1" by auto
    show ?case unfolding idL idR using ind by auto
  qed
qed

definition inverse_base_term :: "('f,'v)term \<Rightarrow> nat \<Rightarrow> ('f \<times> nat,'v)term list"
  where "inverse_base_term l c \<equiv> let hs = [0..< Suc c] in flatten_term_enum (map_term (\<lambda> f. map (\<lambda> h. lift h f) hs) l)"

lemma inverse_base_term: "set (inverse_base_term l c) = 
  {l'. base_term l' = (l :: ('f,'v)term) \<and> (\<forall> f n. (f,n) \<in> funas_term l' \<longrightarrow> height f \<le> c)}"
proof -
  obtain cs where cs: "cs = [0..<Suc c]" by auto
  let ?m = "map_term (\<lambda>f. set (map (\<lambda>h. lift h f) cs))"
  let ?i = "\<lambda> u l. instance_term u (?m l)"
  let ?h = "\<lambda>l. (\<forall> f n. (f,n) \<in> funas_term l \<longrightarrow> height f \<le> c)"
  let ?b = "\<lambda>l' l. base_term l' = l \<and> ?h l'"
  let ?I = "\<lambda>l. {u. ?i u l}"
  let ?B = "\<lambda>l. {u. ?b u l}"
  have "?I l = ?B l" 
  proof (induct l)
    case (Var x)
    {
      fix u
      have "(u \<in> ?I (Var x)) = (u \<in> ?B (Var x))"
        by (cases u, auto)
    }
    thus ?case by auto
  next
    case (Fun f ss) note ind = this
    let ?s = "Fun f ss"
    {
      fix u
      have "(u \<in> (?I ?s)) = (u \<in> (?B ?s))"
      proof (cases u)
        case (Var x)
        thus ?thesis by auto
      next
        case (Fun g ts)
        let ?t = "Fun g ts"
        obtain gg cc where g: "g = (gg,cc)" by (cases g, auto)
        show ?thesis 
        proof (cases "length ts = length ss \<and> gg = f \<and> cc \<le> c")
          case False
          thus ?thesis unfolding Fun g cs by auto
        next
          case True
          hence len: "length ts = length ss" and u: "u = Fun (f,cc) ts"
            and cc: "cc \<le> c" and cc2: "cc \<in> set cs" using Fun g unfolding cs by auto
          {
            fix i 
            assume "i < length ss"
            hence "ss ! i \<in> set ss" by auto
            from ind[OF this] 
            have "?i (ts ! i) (ss ! i) = ?b (ts ! i) (ss ! i)" by auto
          } note ind = this          
          have "?i u ?s = ?b u ?s" unfolding u map_term.simps funas_term.simps
            set_map set_conv_nth[of ts]
            by (simp add: len cc cc2, unfold map_nth_eq_conv[OF len, of base_term] , insert ind, simp, blast)
          thus ?thesis by auto
        qed
      qed
    }
    thus ?case by simp
  qed
  thus ?thesis   unfolding inverse_base_term_def Let_def flatten_term_enum cs map_term_comp o_def .
qed

definition inverse_base_term_filter :: "(('f \<times> nat,'v)term \<Rightarrow> bool) \<Rightarrow> ('f,'v)term \<Rightarrow> nat \<Rightarrow> ('f \<times> nat,'v)term list"
  where "inverse_base_term_filter filt l c \<equiv> let hs = [0..< Suc c] in flatten_term_enum_filter filt (map_term (\<lambda> f. map (\<lambda> h. lift h f) hs) l)"

lemma inverse_base_term_filter: "set (inverse_base_term_filter filt l c) = 
  {l'. base_term l' = (l :: ('f,'v)term) \<and> (\<forall> u \<unlhd> l'. filt u) \<and> (\<forall> f n. (f,n) \<in> funas_term l' \<longrightarrow> height f \<le> c)}"
proof -
  obtain cs where cs: "cs = [0..<Suc c]" by auto
  let ?m = "map_term (\<lambda>f. set (map (\<lambda>h. lift h f) cs))"
  let ?f = "\<lambda> u. (\<forall> v \<unlhd> u. filt v)"
  let ?i' = "\<lambda> u l. instance_term u (?m l)"
  let ?i = "\<lambda> u l. ?i' u l \<and> ?f u"
  let ?h = "\<lambda>l. (\<forall> f n. (f,n) \<in> funas_term l \<longrightarrow> height f \<le> c)"
  let ?b' = "\<lambda>l' l. base_term l' = l \<and> ?h l'"
  let ?b = "\<lambda>l' l. base_term l' = l \<and> ?f l' \<and> ?h l'"
  let ?I = "\<lambda>l. {u. ?i u l}"
  let ?B = "\<lambda>l. {u. ?b u l}"
  { 
    fix u
    have "?i u l = ?b u l"
    proof (induct l arbitrary: u)
      case (Var x u)
      show ?case by (cases u, auto)
    next
      case (Fun f ss u) note ind = this
      let ?s = "Fun f ss"
      show ?case
      proof (cases u)
        case (Var x)
        thus ?thesis by auto
      next
        case (Fun g ts)
        let ?t = "Fun g ts"
        obtain gg cc where g: "g = (gg,cc)" by (cases g, auto)
        show ?thesis 
        proof (cases "length ts = length ss \<and> gg = f \<and> cc \<le> c \<and> ?f ?t")
          case False
          thus ?thesis unfolding Fun g cs by auto
        next
          case True
          hence len: "length ts = length ss" and u: "u = Fun (f,cc) ts"
            and cc: "cc \<le> c" and cc2: "cc \<in> set cs" and ff: "?f (Fun (f,cc) ts)" using Fun g unfolding cs by auto
          {
            fix i 
            assume i: "i < length ss"
            with len have "ts ! i \<in> set ts" by auto
            hence sub: "Fun (f,cc) ts \<unrhd> ts ! i" by auto
            have ff: "?f (ts ! i)"
            proof(intro allI impI)
              fix v
              assume "ts ! i \<unrhd> v" 
              with sub have "Fun (f,cc) ts \<unrhd> v" by (rule supteq_trans)
              with ff show "filt v"  by auto
            qed
            from i have "ss ! i \<in> set ss" by auto
            from ind[OF this, of "ts ! i"] ff 
            have "?i' (ts ! i) (ss ! i) = ?b' (ts ! i) (ss ! i)"              
              by auto
          } note ind = this
          show "?i u ?s = ?b u ?s" unfolding u map_term.simps funas_term.simps
            set_map set_conv_nth[of ts]
            by (simp add: len cc cc2, unfold map_nth_eq_conv[OF len, of base_term] , insert ind, simp, blast)
        qed
      qed
    qed
  }
  thus ?thesis   unfolding inverse_base_term_filter_def Let_def flatten_term_enum_filter cs map_term_comp o_def by auto
qed
    
(* TODO: one can improve the efficiency of the the generator even further, 
   if partial results of the filter can be reused *)
definition cover_bound_list_filter :: "(('f \<times> nat,'v)term \<Rightarrow> bool) \<Rightarrow> (('f,'v)rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool) \<Rightarrow> nat \<Rightarrow> ('f,'v)rules \<Rightarrow> ('f \<times> nat,'v)rule list"
  where "cover_bound_list_filter filt ff c R \<equiv> concat (map (\<lambda>(l,r). 
            (let ee = ff (l,r) in map
                   (\<lambda> l'. ((l', lift_term 
                        (Suc (min_list (map height (sym_collect (\<lambda> t'. ee (base_term t')) l')))) r)))
              (inverse_base_term_filter filt l c)))
    R)"


lemma cover_bound_list_filter: "set (cover_bound_list_filter ff e c (R :: ('f,'v)rules)) = (cover_bound e c (set R) \<inter> {(l,r). (\<forall> u \<unlhd> l. ff u)})" (is "?L = ?R")
proof -
  obtain h where h: "h = (\<lambda> l' l r. Suc (min_list (map height (sym_collect (\<lambda>t'. e (l,r) (base_term t')) l'))))" by auto
  obtain RR where RR: "cover_bound e c (set R) \<inter> {(l,r). (\<forall> u \<unlhd> l. ff u)} = RR" by auto
  let ?f = "\<lambda>l. (\<forall> u \<unlhd> l. ff u)"
  {
    fix l' r' 
    assume l'r': "(l',r') \<in> ?L"
    from l'r'[unfolded cover_bound_list_filter_def Let_def]
    obtain l r where lr: "(l,r) \<in> set R" and
      l'r': "(l',r') \<in> set (map (\<lambda>l'. (l', lift_term (h l' l r) r)) (inverse_base_term_filter ff l c))"
      unfolding h set_concat  set_map by force
    hence 
      r': "r' = lift_term (h l' l r) r"
      and l': "l' \<in> set (inverse_base_term_filter ff l c)"
      by auto
    from l'[unfolded inverse_base_term_filter]
    have l': "base_term l' = l"
      and c: "\<forall> f n. (f,n) \<in> funas_term l' \<longrightarrow> height f \<le> c" 
      and f: "?f l'" by auto
    from lr r' l' c have cover: "(l',r') \<in> cover e (set R)" unfolding cover_def h by auto
    have "(l',r') \<in> ?R" using c cover f unfolding cover_bound_def by auto
  }
  moreover
  {
    fix l' r' 
    assume "(l',r') \<in> ?R"
    hence cover: "(l',r') \<in> cover e (set R)" 
      and c: "\<forall> f n. (f,n) \<in> funas_term l' \<longrightarrow> height f \<le> c"      
      and f: "?f l'"
      unfolding cover_bound_def by auto    
    from cover[unfolded cover_def]
    obtain l r where
      lr: "(l,r) \<in> set R"
      and r': "r' = lift_term (h l' l r) r"
      and l': "base_term l' = l"
      unfolding h by blast      
    from c l' f have l': "l' \<in> set (inverse_base_term_filter ff l c)" 
      unfolding inverse_base_term_filter by auto
    have "(l',r') \<in> ?L"
      by (unfold cover_bound_list_filter_def Let_def set_concat set_map, rule, rule, rule lr, unfold set_map r' h, rule, rule refl, rule l')
  }
  ultimately show ?thesis unfolding RR by force
qed

lemma remove_compatible_rules: 
  assumes compat: "ta_trs_compatible TA (cover_bound e c R \<inter> {(l,r). ff l r})"
  and filter: "\<And> l r. \<not> ff l r \<Longrightarrow> ta_rule_compatible TA (l,r)"
  shows "ta_trs_compatible TA (cover_bound e c R)"
  using assms unfolding ta_trs_compatible_def by blast

datatype boundstype = Roof | Match

fun boundstype_fun :: "boundstype \<Rightarrow> ('f,'v) rule \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
  where "boundstype_fun Roof = Matchbounds.roof"
      | "boundstype_fun Match = Matchbounds.match"

fun bounds_condition :: "boundstype \<Rightarrow> ('f :: show ,'v :: show)rules \<Rightarrow> shows check"
where "bounds_condition Roof _ = OK"
   |  "bounds_condition Match R = (check_all (\<lambda>(l,r). linear_term r) R
              <+? (\<lambda> (l,r). shows ''rule '' +@+ shows_rule (l,r) +@+ shows '' is not right-linear ''))"

(* currently we use right linearity to ensure non-duplicatingness, since
   linearity we can already check and by demanding left-linearity anyway,
   this is no restricting. As soon as we allow bounds for non-ll-systems,
   one should replace the rl-condition for match-bounds by nd *)


lemma bounds_condition: assumes ok: "isOK(bounds_condition type R)"
  and wf: "wf_trs (set R)"
  shows "locally_terminating (cover (boundstype_fun type) (set R))"
proof (cases type)
  case Roof
  show ?thesis unfolding Roof
    using roof_locally_SN[OF wf]
    by auto
next
  case Match
  show ?thesis unfolding Match
  proof (unfold boundstype_fun.simps, rule match_lSN[OF wf])
    fix l r
    assume "(l,r) \<in> set R"
    with ok[unfolded Match] wf[unfolded wf_trs_def] have 
      "linear_term r" and "vars_term r \<subseteq> vars_term l" 
      by auto
    from right_linear_imp_non_duplicating[OF this]
    show "vars_term_ms r \<le> vars_term_ms l" .
  qed
qed
 
datatype ('f,'v)bounds_info = Bounds_Info boundstype nat "'v list" "('v, 'f \<times> nat) tree_automaton"


(* currently type of variables of TRS and states of q must coincide, this can be generalized if there is demand *)
fun check_bounds :: "('f :: {show,linorder},'q :: {show,linorder})bounds_info \<Rightarrow> ('f,'q)rules \<Rightarrow> shows check" where
  "check_bounds (Bounds_Info type c qfin preTA) R = (do {
     let TA = generate_ta preTA;
     check_wf_trs R;
     check_left_linear_trs R;
     bounds_condition type R;
     check (list_ex (\<lambda> q. rs_memb q (ta_rqs_set TA)) qfin) (shows ''did not find mentioned final state in TA'');
     check_all (\<lambda> (f,n). height f \<le> c) (map fst (rm_to_list (ta_rules_impl TA)))
        <+? (\<lambda> (f,n). shows f +@+ shows '' is symbol in TA with height larger than c = '' +@+ shows c);
     ta_contains_impl (map (\<lambda>(f,n). (lift 0 f,n)) (funas_trs_list R)) TA qfin
        <+? (\<lambda> e. shows ''it could not be guaranteed that lift0(T(Sigma)) is accepted by TA'' +@+ shows_nl +@+
              (case e of Inl q \<Rightarrow> shows ''state '' +@+ shows q +@+ shows '' is not final''
                       | Inr fqs \<Rightarrow> shows ''there is no transition from '' +@+ shows fqs +@+ shows '' to a final state''));
     ta_trs_compatible_eff_list TA (cover_bound_list_filter (\<lambda> l. \<not> ta_rule_compatible_heuristic TA l) (boundstype_fun type) c R)
        <+? (\<lambda> (lr,lrq,q). shows ''TA is not compatible with R'' +@+ shows_nl 
             +@+ shows ''for rule '' +@+ shows_rule lr +@+ shows_nl
             +@+ shows ''which is instantiated by states to '' +@+ shows_rule lrq +@+ shows_nl
             +@+ shows ''the state '' +@+ shows q +@+ shows '' is only reachable from the lhs'' +@+ shows_nl)
   })"

lemma check_bounds: assumes ok: "isOK(check_bounds info R)"
  shows "SN (rstep (set R))"
proof -
  obtain type c qfin preTA where info: "info = Bounds_Info type c qfin preTA"
    by (cases info, blast+)
  note ok = ok[unfolded info]
  obtain TA where TA: "generate_ta preTA = TA" by auto
  let ?filt = "\<lambda> l. \<not> ta_rule_compatible_heuristic TA l"
  let ?sfilt = "\<lambda> l. (\<forall> u \<unlhd> l. ?filt u)"
  note ok = ok[unfolded check_bounds.simps Let_def TA, simplified]
  interpret ta_inv TA using generate_ta[of preTA] unfolding TA .
  from ok[unfolded list_ex_iff] have 
    wf: "wf_trs (set R)"
    and ll: "left_linear_trs (set R)"
    and qfin: "\<exists> q \<in> set qfin. rs_memb q (ta_rqs_set TA)"
    and c: "\<And> fn. fn \<in> set (rm_to_list (ta_rules_impl TA)) \<Longrightarrow> case fst fn of (f,n) \<Rightarrow> height f \<le> c"
    and cond: "isOK(bounds_condition type R)"
    and contain: "isOK(ta_contains_impl (map (\<lambda>(f,n). (lift 0 f,n)) (funas_trs_list R)) TA qfin)"
    and compat: "isOK(ta_trs_compatible_eff_list TA (cover_bound_list_filter ?filt (boundstype_fun type) c R))" by auto
  from qfin have qfin: "\<exists> q \<in> set qfin. q \<in> ta_rqs TA'" using ta_rqs_set by (auto simp: rs_correct)
  let ?C = "cover_bound (boundstype_fun type) c (set R)"
  let ?Cf = "cover_bound (boundstype_fun type) c (set R) \<inter> {(l,r). ?sfilt l}"
  let ?Rf = "?C \<inter> {(l,r). ?sfilt l}"
  have C: "?Rf = ?Cf" by force
  {
    fix f n
    assume "(f,n) \<in> ta_syms TA'"
    then obtain rule where rule: "rule \<in> ta_rules TA'" and n: "rsym rule = (f,n)"
      unfolding ta_syms_def by auto        
    let ?crule = "conv_ta_rule (ta_epss_impl TA) rule"
    from rule have crule: "?crule \<in> conv_ta_rule (ta_epss_impl TA) ` ta_rules TA'" by auto 
    from n have n: "rsym_impl ?crule = (f,n)" by (cases rule, auto)
    let ?rm = "ta_rules_impl TA"
    from rm_set_lookup3[OF crule, unfolded n]
    obtain rls where "rm_\<alpha> ?rm (f,n) = Some rls" by force
    hence "map_of (rm_to_list ?rm) (f,n) = Some rls" by (auto simp: rm_correct)
    from map_of_SomeD[OF this]
    have "((f,n),rls) \<in> set (rm_to_list ?rm)" .
    from c[OF this] have "height f \<le> c" by simp
  } note c = this  
  have "ta_trs_compatible TA' ?Rf"    
  proof (rule ta_trs_compatible_eff_list[OF _ compat, unfolded cover_bound_list_filter])
    fix l r
    assume "(l,r) \<in> ?Rf"
    hence "(l,r) \<in> cover (boundstype_fun type) (set R)" unfolding cover_bound_def by auto
    from cover_var_condition[OF _ this] wf[unfolded wf_trs_def]
    show "vars_term r \<subseteq> vars_term l" by auto
  qed 
  hence "ta_trs_compatible TA' ?Cf" .
  hence compat: "ta_trs_compatible TA' ?C"
  proof (rule remove_compatible_rules, clarify)
    fix l r u
    assume compat: "ta_rule_compatible_heuristic TA u"
    and ncompat: "\<not> ta_rule_compatible TA' (l,r)"
    and sub: "l \<unrhd> u"
    from ta_rule_compatible_heuristic_subteq[OF compat sub] ncompat
    show False by auto
  qed
  obtain fin rls eps where preTA: "preTA = Tree_Automaton fin rls eps" by (cases preTA, auto)
  from generate_ta_rules[of fin rls eps] have "ta_rules (ta_of TA) = set rls" using TA preTA by auto
  hence fin: "finite (ta_rules (ta_of TA))" by auto
  show ?thesis
    by (rule e_bounds_impl[OF ll wf fin bounds_condition[OF cond wf] compat c _ qfin], insert ta_contains_impl[OF contain], auto) 
qed        

(* silently convert weak to strict rules *)
definition
  bounds_tt ::
    "('tp, 'f, 'q) tp_ops \<Rightarrow>
    ('f::{show,linorder}, 'q::{show,linorder}) bounds_info \<Rightarrow>
    'tp \<Rightarrow> shows check"
where
  "bounds_tt I info tp \<equiv> do {
      let r = tp_ops.rules I tp;
      check_bounds info r
   }"

lemma bounds_tt:
  assumes "tp_spec I" and ok: "isOK (bounds_tt I info tp)"
  shows "SN_qrel (tp_ops.qreltrs I tp)"
proof -
  interpret tp_spec I by fact
  from check_bounds[OF ok[unfolded bounds_tt_def, simplified]]
    have SN: "SN_qrel ({}, set (R tp) \<union> set (Rw tp), {})" by simp
  show ?thesis unfolding qreltrs_sound
    by (rule SN_qrel_mono[OF _ _ _ SN]) simp_all
qed

end
